#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/if.h>
#include <sys/time.h>
#include <stdarg.h>

#include "monarch.h"

static int verbosity = 0;

char *renderIP(in_addr_t ip, char *buf)
{
  sprintf(buf, "%d.%d.%d.%d", (ip)>>24, ((ip)>>16)&0xFF, ((ip)>>8)&0xFF, ip & 0xFF);
  return buf;
}

char *renderAddress(struct sockaddr_in *addr, char *buf)
{
  assert(addr->sin_family == AF_INET);
  if (addr->sin_addr.s_addr)
    sprintf(buf, "%d.%d.%d.%d", addr->sin_addr.s_addr & 0xFF, ((addr->sin_addr.s_addr)>>8)&0xFF, ((addr->sin_addr.s_addr)>>16)&0xFF, (addr->sin_addr.s_addr)>>24);
  else
    strcpy(buf, "*");
  
  char *portPart = &buf[strlen(buf)];
  if (addr->sin_port)
    sprintf(portPart, ":%d", ntohs(addr->sin_port));
  else
    strcpy(portPart, ":*");
    
  return buf;
}

int parseAddressAndPort(const char *string, struct sockaddr_in *returnValue)
{
  if (!string)
    return 0;
    
  char buf[strlen(string)+1];
  strcpy(buf, string);
  
  char *hostPart = strtok(buf, ":");
  char *portPart = hostPart ? strtok(NULL, "\r") : NULL;
  
  struct hostent *ent = gethostbyname(hostPart);
  if (!ent)
    return 0;
    
  int port = 0;
  if (portPart) {
    char *endptr = NULL;
    port = strtol(portPart, &endptr, 10);
    if (endptr && *endptr)
      return 0;
  }
    
  memset(returnValue, 0, sizeof(*returnValue));
  returnValue->sin_addr.s_addr = *(in_addr_t*)ent->h_addr_list[0];
  returnValue->sin_port = htons(port);
  returnValue->sin_family = AF_INET;
  
  return 1;
}


in_addr_t getIP(const char *ifName)
{
  int fd = socket(AF_INET, SOCK_DGRAM, 0);
  struct ifreq ifr;
  
  strcpy(ifr.ifr_name, ifName);
  if (ioctl(fd, SIOCGIFADDR, &ifr) != 0)
    panic(ERR_INTERNAL, "Cannot get IP of interface '%s'", ifName);
  close(fd);
  
  struct sockaddr_in *saddr = (struct sockaddr_in*) &ifr.ifr_addr;
  return saddr->sin_addr.s_addr;
}

void findUnusedPorts(in_addr_t ip, uint16 *ports, int num)
{
  int sock[num];
  for (int i=0; i<num; i++) {
    struct sockaddr_in addr;
    addr.sin_family = AF_INET;
    addr.sin_port = 0;
    addr.sin_addr.s_addr = ip;
    sock[i] = socket(AF_INET, SOCK_STREAM, 0);
    assert(sock[i] >= 0);
    int ret = bind(sock[i], (struct sockaddr*)&addr, sizeof(addr));
    assert(ret >= 0);
    socklen_t len = sizeof(addr);
    ret = getsockname(sock[i], (struct sockaddr*)&addr, &len);
    assert(ret >= 0);
    ports[i] = addr.sin_port;
    int yes = 1;
    ret = setsockopt(sock[i], SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
    assert(ret == 0);
  }
  
  for (int i=0; i<num; i++)
    close(sock[i]);
}

void setVerbosity(int level)
{
  verbosity = level;
}

int getVerbosity()
{
  return verbosity;
}

void verbose(int level, const char *format, ...)
{
  char buffer[256];
  va_list ap;
  va_start(ap, format);
  vsnprintf(buffer, sizeof(buffer), format, ap);
  if (level <= verbosity) {
    fprintf(stderr, "%s\n", buffer);
    fflush(stderr);
  }
  va_end (ap);
}
