#include <stdlib.h>
#include <string.h>

#include "monarch.h"
#include "probes.h"
#include "proxies.h"

#define dprintf(a...) do { } while (0)
//#define dprintf(a...) printf(a)

IcmpTimestampRequestPolicy::IcmpTimestampRequestPolicy() : ProbePolicy()
{
  this->nextSequenceNo = 0;
  this->packetTable = new Hashtable();
}

void IcmpTimestampRequestPolicy::receivePacketFromSender(Packet *packet)
{
  dprintf("Receive packet from sender\n");
  int thisSequenceNo = nextSequenceNo++;
  packetTable->put(thisSequenceNo, new Packet(packet));
  proxy->sendPacketToRemote(makeIcmpTimestampRequest(packet, thisSequenceNo));
  probeTrace->probeSent(thisSequenceNo);
}

void IcmpTimestampRequestPolicy::receivePacketFromRemote(Packet *packet)
{
  dprintf("Receive packet from remote\n");
  struct IcmpTimestampRequest *tsr = (struct IcmpTimestampRequest*) &(packet->getBuffer()[20]);
  int thisIndex = ntohs(tsr->sequenceno);
  Packet *savedPacket = (Packet*) packetTable->get(thisIndex);
  
  if (!savedPacket)
    panic(ERR_INTERNAL, "TimestampRequestPolicy: No saved packet found for incoming TimestampReply #%d", thisIndex);
  
  probeTrace->responseReceived(thisIndex, packet->getIpId());
  proxy->forwardPacketToReceiver(savedPacket);
}

void IcmpTimestampRequestPolicy::receivePacketFromReceiver(Packet *packet)
{
  dprintf("Receive packet from receiver\n");
  proxy->forwardPacketToSender(packet);
}

Packet *IcmpTimestampRequestPolicy::makeIcmpTimestampRequest(Packet *packet, int sequenceNo)
{
  Packet *request = new Packet();
  
  struct IcmpTimestampRequest *tsr = (struct IcmpTimestampRequest*) &(request->getBuffer()[20]);
  struct IpHeader *ip = (struct IpHeader*) request->getBuffer();
  uint16 iplen = sizeof(*ip) + sizeof(*tsr);
  request->setLength(iplen);

  memset(request->getBuffer(), 0, request->getBufferLen());
  ip->ip_proto = IPPROTO_ICMP; 
  ip->ip_ttl = 60; 
  ip->ip_tos = 0; 
  ip->ip_vhl = 0x45; 
  ip->ip_off = htons(IP_DF); 
  ip->ip_len = htons(iplen);
  ip->ip_id = 0;

  tsr->type = 13;
  tsr->code = 0;
  tsr->checksum = 0;
  tsr->identifier = ~htons(sequenceNo);
  tsr->sequenceno = htons(sequenceNo);
  tsr->originateTime = 0;
  tsr->receiveTime = 0;
  tsr->transmitTime = 0;
  tsr->checksum = htons(Packet::checksum((uint8*)tsr, sizeof(*tsr)));

  return request;
}
