#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "monarch.h"
#include "probes.h"
#include "proxies.h"

TcpAckPolicy::TcpAckPolicy() : ProbePolicy()
{
  this->packetsReceivedFromSender = 0;
  this->packetsReceivedFromRemote = 0;
  this->packetsSentToRemote = 0;
  this->packetTable = new Hashtable();
  this->nextAckIndex = 0;
}

void TcpAckPolicy::receivePacketFromSender(Packet *packet)
{
  packetsReceivedFromSender ++;

  /* Check whether we the remote host does not answer with RESETs. A
   * simple check is to test whether the number of packets already sent
   * is large enough */
   
  if (!packetsReceivedFromRemote && (packetsSentToRemote > 3))
    panic(ERR_TIMEOUT, "Remote host not resetting... Timeout");

  /* Remember the packet; it is later forwarded to the receiver */

  int thisAckIndex = nextAckIndex ++;  
  packetTable->put(thisAckIndex, new Packet(packet));
  
  /* Create and send the packet */

  Packet *ack = makeAck(packet, firstAckSeqNo + thisAckIndex);

  /* Send probe packet to remote */
  
  proxy->sendPacketToRemote(ack);
  probeTrace->probeSent(thisAckIndex);
  packetsSentToRemote ++;
}

void TcpAckPolicy::receivePacketFromRemote(Packet *packet)
{
  /* Find out the sequence number */
  
  int thisAckIndex = packet->getTcpSequenceNo() - firstAckSeqNo;  
  if (thisAckIndex < 0) {
    fprintf(stderr, "Unknown index for incoming RST: #%d\n", thisAckIndex);
    return;
  }
  
  probeTrace->responseReceived(thisAckIndex, packet->getIpId());
  
  /* Look up the corresponding TCP packet */

  Packet *savedPacket = (Packet*) packetTable->get(thisAckIndex);
  if (!savedPacket) {
    packet->dump(stderr);
    panic(ERR_INTERNAL, "No saved packet found for incoming RST #%d", thisAckIndex);
  }
  
  packetsReceivedFromRemote ++;

  proxy->forwardPacketToReceiver(savedPacket);
}

void TcpAckPolicy::receivePacketFromReceiver(Packet *packet)
{
  proxy->forwardPacketToSender(packet);
}

Packet *TcpAckPolicy::makeAck(Packet *original, uint32 idx) 
{
  Packet *ack = new Packet(original);

  struct IpHeader *ip = (struct IpHeader*) ack->getBuffer();
  struct TcpHeader *tcp = (struct TcpHeader*) &(ack->getBuffer()[20]);

  uint16 iplen = ntohs(ip->ip_len);
  assert(iplen >= 40);

  /* Set the tcp header length correct */
  int tcpHeaderLen;
  if (ack->isTcpPacket())
    /* We just send on all the options that we found in the original packet */
    tcpHeaderLen = tcp->tcp_hl;
  else if (ack->isUdpPacket())
    /* No options for transformed UDP packets */
    tcpHeaderLen = 0x50;
  else
    assert(false);

  memset(ack->getBuffer(), 0, sizeof (struct IpHeader));
  ip->ip_proto = IPPROTO_TCP; 
  ip->ip_ttl = 60; 
  ip->ip_tos = 0; 
  ip->ip_vhl = 0x45; 
  ip->ip_off = htons(IP_DF); 
  ip->ip_len = htons(iplen); 

  tcp->tcp_flags = Packet::FLAG_ACK;
  tcp->tcp_seq = htonl(idx);
  tcp->tcp_ack = htonl(idx);
  tcp->tcp_hl = tcpHeaderLen;
  tcp->tcp_win = 1000;
  tcp->tcp_urp = 0;

  return ack;
}
