#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>

#include "monarch.h"
#include "traces.h"

KernelTrace::KernelTrace()
{
  firstBlock = (struct snapshotBlock*) malloc(sizeof(struct snapshotBlock));
  assert(firstBlock != NULL);
  
  firstBlock->snapshotsHere = 0;
  firstBlock->nextBlock = NULL;
  lastBlock = firstBlock;
  firstComment = NULL;
  lastComment = NULL;
  outFileName[0] = 0;
}

KernelTrace::~KernelTrace()
{
  while (firstBlock != NULL) {
    struct snapshotBlock *nextBlock = firstBlock->nextBlock;
    free(firstBlock);
    firstBlock = nextBlock;
  }
  
  while (firstComment != NULL) {
    struct commentRecord *nextComment = firstComment->next;
    free(firstComment);
    firstComment = nextComment;
  }
}

void KernelTrace::addComment(const char *format, ...)
{
  char buffer[256];
  va_list ap;
  va_start(ap, format);
  vsnprintf(buffer, sizeof(buffer), format, ap);
  struct commentRecord *comment = (struct commentRecord*) malloc(sizeof(commentRecord));
  assert(comment != NULL);
  comment->next = NULL;
  strncpy(comment->text, buffer, sizeof(comment->text));
  
  if (lastComment) 
    lastComment->next = comment;
  else
    firstComment = comment;
  lastComment = comment;
  
  va_end (ap);
}

void KernelTrace::addSnapshot(struct timeval *time, struct tcp_info *state)
{
  if (lastBlock->snapshotsHere >= snapshotsPerBlock) {
    assert(lastBlock->nextBlock == NULL);
    lastBlock->nextBlock = (struct snapshotBlock*) malloc(sizeof(struct snapshotBlock));
    assert(lastBlock->nextBlock != NULL);
    lastBlock = lastBlock->nextBlock;
    lastBlock->snapshotsHere = 0;
    lastBlock->nextBlock = NULL;
  }
  
  int thisIndex = lastBlock->snapshotsHere ++;
  lastBlock->snapshot[thisIndex].state = *state;
  lastBlock->snapshot[thisIndex].time = *time;
}

void KernelTrace::writeSnapshot(struct snapshotRecord *record, FILE *stream)
{
  fprintf(stream, "%d.%06d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
    (int)record->time.tv_sec, (int)record->time.tv_usec, 
    record->state.tcpi_state, record->state.tcpi_ca_state, record->state.tcpi_retransmits,
    record->state.tcpi_probes, record->state.tcpi_backoff, record->state.tcpi_options,
    record->state.tcpi_snd_wscale, record->state.tcpi_rcv_wscale,
    record->state.tcpi_rto, record->state.tcpi_ato, record->state.tcpi_snd_mss,
    record->state.tcpi_rcv_mss, record->state.tcpi_unacked, record->state.tcpi_sacked,
    record->state.tcpi_lost, record->state.tcpi_retrans, record->state.tcpi_fackets,
    record->state.tcpi_last_data_sent, record->state.tcpi_last_ack_sent, record->state.tcpi_last_data_recv,
    record->state.tcpi_last_ack_recv, record->state.tcpi_pmtu, record->state.tcpi_rcv_ssthresh,
    record->state.tcpi_rtt, record->state.tcpi_rttvar, record->state.tcpi_snd_ssthresh,
    record->state.tcpi_snd_cwnd, record->state.tcpi_advmss, record->state.tcpi_reordering
  );
}

void KernelTrace::finalize()
{
  FILE *stream = fopen(outFileName, "w+");
  if (!stream)
    panic(ERR_FILEIO, "Cannot write trace to '%s'", outFileName);
  
  struct commentRecord *currentComment = firstComment;
  while (currentComment) {
    fprintf(stream, "# %s\n", currentComment->text);
    currentComment = currentComment->next;
  }
  
  struct snapshotBlock *currentBlock = firstBlock;
  while (currentBlock) {
    for (int i=0; i<currentBlock->snapshotsHere; i++)
      writeSnapshot(&currentBlock->snapshot[i], stream);
    currentBlock = currentBlock->nextBlock;
  }
  
  fclose(stream);
}
